# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""Drawing classes ShapeBase, LineShape, PolylineShape, PolygonShape,
RectangleShape, EllipseShape, ConnectorShape and DrawGroup.
"""

from __future__ import annotations

from typing import Any

from .element import Element, PropDef, PropDefBool, register_element_class
from .frame import AnchorMix, PosMix, SizeMix, ZMix
from .mixin_list import ListMixin
from .svg import SvgMixin


class ShapeBase(ListMixin, AnchorMix, SvgMixin, ZMix, Element):
    """Base class for all drawing shapes.

    This class provides common properties and methods for various shapes
    like lines, rectangles, ellipses, and connectors. It integrates sizing
    functionalities through mixins.
    """

    _tag = "draw:shape-odfdo-notodf"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "draw:name"),
        PropDef("draw_id", "draw:id"),
        PropDef("layer", "draw:layer"),
        PropDef("presentation_class", "presentation:class-names"),
        PropDef("presentation_style", "presentation:style-name"),
        PropDef("style", "draw:style-name"),
        PropDef("text_style", "draw:text-style-name"),
        PropDef("caption_id", "draw:caption-id"),
        PropDef("class_names", "draw:class-names"),
        PropDef("transform", "draw:transform"),
        PropDef("end_cell_address", "table:end-cell-address"),
        PropDef("end_x", "table:end-x"),
        PropDef("end_y", "table:end-y"),
        PropDefBool("table_background", "table:table-background", False),
        PropDef("xml_id", "xml:id"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ShapeBase element.

        Args:
            name: Name of the graphical element.
            style: The style name for the shape.
            text_style: The text style name for the shape.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the shape.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            if style:
                self.style = style
            if text_style:
                self.text_style = text_style
            if draw_id:
                self.draw_id = draw_id
            if layer:
                self.layer = layer
            if presentation_class:
                self.presentation_class = presentation_class
            if presentation_style:
                self.presentation_style = presentation_style
            if caption_id:
                self.caption_id = caption_id
            if class_names:
                self.class_names = class_names
            if transform:
                self.transform = transform
            if z_index is not None:
                self.z_index = z_index
            if end_cell_address:
                self.end_cell_address = end_cell_address
            if end_x:
                self.end_x = end_x
            if end_y:
                self.end_y = end_y
            if table_background is not None:
                self.table_background = table_background
            if anchor_type:
                self.anchor_type = anchor_type
            if anchor_page is not None:
                self.anchor_page = anchor_page
            if xml_id:
                self.xml_id = xml_id

    def get_formatted_text(self, context: dict | None = None) -> str:
        """Return the formatted text content of the shape.

        Args:
            context: A dictionary of context variables.

        Returns:
            str: The formatted text content.
        """
        result: list[str] = []
        for child in self.children:
            result.append(child.get_formatted_text(context))
        result.append("\n")
        return "".join(result)


ShapeBase._define_attribut_property()


class AngleMix(Element):
    """Kind, start_angle and end_angle for circle and ellipse.

    Kind value can be: 'full', 'section', 'cut' or 'arc'. Default is 'full'.
    """

    _tag = "draw:anglemix-odfdo-notodf"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("start_angle", "draw:start-angle"),
        PropDef("end_angle", "draw:end-angle"),
    )

    KIND_VALUE_CHOICE = {  # noqa: RUF012
        "full",
        "section",
        "cut",
        "arc",
    }

    @property
    def kind(self) -> str:
        'Get or set the kind, "draw:kind".'
        return self._get_attribute_str_default("draw:kind", "full")

    @kind.setter
    def kind(self, kind: str) -> None:
        if kind not in self.KIND_VALUE_CHOICE:
            raise TypeError(f"'draw:kind' not valid: {kind!r}")
        self._set_attribute_str_default("draw:kind", kind, "full")


class LineShape(ShapeBase):
    """Represents a line shape, "draw:line".

    This shape defines a straight line between two points.
    """

    _tag = "draw:line"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("x1", "svg:x1"),
        PropDef("y1", "svg:y1"),
        PropDef("x2", "svg:x2"),
        PropDef("y2", "svg:y2"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        p1: tuple[str, str] | list[str] | None = None,
        p2: tuple[str, str] | list[str] | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a line shape "draw:line".

        Args:
            name: Name of the graphical element.
            style: The style name for the line.
            text_style: The text style name for the line.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the line.
            p1: The (x1, y1) coordinates of the starting point.
            p2: The (x2, y2) coordinates of the ending point.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            self.p1 = p1
            self.p2 = p2

    @property
    def p1(self) -> tuple[str | None, str | None]:
        "Get or set the (x1, y1) coordinates of the starting point."
        return (self.x1, self.y1)

    @p1.setter
    def p1(self, p1: tuple[str, str] | list[str] | None) -> None:
        if p1 is None:
            self.x1 = None
            self.y1 = None
        else:
            self.x1 = p1[0]
            self.y1 = p1[1]

    @property
    def p2(self) -> tuple[str | None, str | None]:
        "Get or set the (x2, y2) coordinates of the ending point."
        return (self.x2, self.y2)

    @p2.setter
    def p2(self, p2: tuple[str, str] | list[str] | None) -> None:
        if p2 is None:
            self.x2 = None
            self.y2 = None
        else:
            self.x2 = p2[0]
            self.y2 = p2[1]


LineShape._define_attribut_property()


class DrawMeasure(LineShape):
    """Represents a shape that is used to measure distances in drawings,
    "draw:measure".
    """

    _tag = "draw:measure"

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        p1: tuple[str, str] | list[str] | None = None,
        p2: tuple[str, str] | list[str] | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a line shape "draw:line".

        Args:
            name: Name of the graphical element.
            style: The style name for the measure.
            text_style: The text style name for the measure.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the measure.
            p1: The (x1, y1) coordinates of the starting point.
            p2: The (x2, y2) coordinates of the ending point.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "p1": p1,
                "p2": p2,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)


class RectangleShape(PosMix, SizeMix, ShapeBase):
    """Represents a rectangle shape, "draw:rect".

    This shape defines a rectangular area.
    """

    _tag = "draw:rect"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("corner_radius", "draw:corner-radius"),
        PropDef("rx", "svg:rx"),
        PropDef("ry", "svg:ry"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        corner_radius: str | None = None,
        rx: str | None = None,
        ry: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a rectangle shape "draw:rect".

        Args:
            name: Name of the graphical element.
            style: The style name for the rectangle.
            text_style: The text style name for the rectangle.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the rectangle.
            position: The (x, y) coordinates for the rectangle's position.
            size: The (width, height) values for the rectangle's size.
            corner_radius: radius of the circle used to round off the corners.
            rx: x-axis radius of the ellipse used to round off the corners.
            ry: y-axis radius of the ellipse used to round off the corners.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if position:
                self.position = position
            if size:
                self.size = size
            if corner_radius:
                self.corner_radius = corner_radius
            if rx:
                self.rx = rx
            if ry:
                self.ry = ry


RectangleShape._define_attribut_property()


class PolylineShape(PosMix, SizeMix, ShapeBase):
    """Represents a polyline shape, "draw:polyline"."""

    _tag = "draw:polyline"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("points", "draw:points"),
        PropDef("view_box", "svg:viewBox"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        points: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        view_box: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a polyline shape "draw:polyline".

        Args:
            name: Name of the graphical element.
            style: The style name for the polyline.
            text_style: The text style name for the polyline.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the polyline.
            points: The coordinates of the polyline.
            position: The (x, y) coordinates for the polyline's position.
            size: The (width, height) values for the polyline's size.
            view_box: The rectangle in a local coordinates system used by the
                points.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if points:
                self.points = points
            if position:
                self.position = position
            if size:
                self.size = size
            if view_box:
                self.view_box = view_box


PolylineShape._define_attribut_property()


class PolygonShape(PolylineShape):
    """Represents a polygon, "draw:polygon".

    A polygon is a closed set of straight lines."""

    _tag = "draw:polygon"

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        points: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        view_box: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a polygon shape "draw:polygon>".

        Args:
            name: Name of the graphical element.
            style: The style name for the polygon.
            text_style: The text style name for the polygon.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the polygon.
            points: The coordinates of the polygon.
            position: The (x, y) coordinates for the polygon's position.
            size: The (width, height) values for the polygon's size.
            view_box: The rectangle in a local coordinates system used by the
                points.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "points": points,
                "position": position,
                "size": size,
                "view_box": view_box,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)


PolygonShape._define_attribut_property()


class RegularPolygonShape(PosMix, SizeMix, ShapeBase):
    """Represents a regular polygon, "draw:regular-polygon".

    A regular polygon is a polygon that is specified by its number of edges
    (that is equal to the number of its corners), rather than by arbitrary
    points."""

    _tag = "draw:regular-polygon"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDefBool("concave", "draw:concave", False),
        PropDef("sharpness", "draw:sharpness"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        corners: int | None = None,
        concave: bool | None = None,
        sharpness: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a regular polygon shape "draw:regular-polygon".

        Args:
            name: Name of the graphical element.
            style: The style name for the polygon.
            text_style: The text style name for the polygon.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the polygon.
            corners: The number of polygon corners.
            concave: Whether a regular polygon is convex or concave.
            sharpness: The radius of the ellipse on which inner polygon
                corners are located for concave polygon.
            position: The (x, y) coordinates for the polygon's position.
            size: The (width, height) values for the polygon's size.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if corners:
                self.corners = corners
            if concave is not None:
                self.concave = concave
            if sharpness:
                self.sharpness = sharpness
            if position:
                self.position = position
            if size:
                self.size = size

    @property
    def corners(self) -> int | None:
        """Get or set the number of polygon corners.

        type : int or None
        """
        corners = self.get_attribute("draw:corners")
        if corners is None:
            return None
        return int(corners)

    @corners.setter
    def corners(self, corners: int | None) -> None:
        self._set_attribute_int("draw:corners", corners)


RegularPolygonShape._define_attribut_property()


class DrawPath(PosMix, SizeMix, ShapeBase):
    """Represents a path, "draw:path".

    A path is a shape with a user-defined outline. The outline is defined by
    the svg:d attribute.
    """

    _tag = "draw:path"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("svg_d", "svg:d"),
        PropDef("view_box", "svg:viewBox"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        svg_d: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        view_box: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a path shape "draw:path".

        Args:
            name: Name of the graphical element.
            style: The style name for the path.
            text_style: The text style name for the path.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the path.
            svg_d: A path data.
            position: The (x, y) coordinates for the path's position.
            size: The (width, height) values for the path's size.
            view_box: The rectangle in a local coordinates system used by the
                points.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if svg_d:
                self.svg_d = svg_d
            if position:
                self.position = position
            if size:
                self.size = size
            if view_box:
                self.view_box = view_box


DrawPath._define_attribut_property()


class DrawCaption(PosMix, SizeMix, ShapeBase):
    """Represents a description attached to a fixed point, "draw:caption".

    It consists of rectangular drawing shape with an additional set of
    connected lines that connect the rectangle with the fixed point.
    """

    _tag = "draw:caption"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("caption_point_x", "draw:caption-point-x"),
        PropDef("caption_point_y", "draw:caption-point-y"),
        PropDef("corner_radius", "draw:corner-radius"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        caption_point: tuple[str, str] | list[str] | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        corner_radius: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a draw caption "draw:caption".

        Args:
            name: Name of the graphical element.
            style: The style name for the path.
            text_style: The text style name for the path.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the path.
            caption_point: The position of a point that is captioned. A set
                of lines is rendered to that point from the caption area.
            position: The (x, y) coordinates for the path's position.
            size: The (width, height) values for the path's size.
            corner_radius: radius of the circle used to round off the corners.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if position:
                self.position = position
            if size:
                self.size = size
            if corner_radius:
                self.corner_radius = corner_radius
            if caption_point:
                self.caption_point = caption_point

    @property
    def caption_point(self) -> tuple[str | None, str | None]:
        """Get or set the (caption_point_x, caption_point_y) coordinates of the
        caption point."""
        return (self.caption_point_x, self.caption_point_y)

    @caption_point.setter
    def caption_point(self, caption_point: tuple[str, str] | list[str] | None) -> None:
        if caption_point is None:
            self.caption_point_x = None
            self.caption_point_y = None
        else:
            self.caption_point_x = caption_point[0]
            self.caption_point_y = caption_point[1]


DrawCaption._define_attribut_property()


class EllipseShape(AngleMix, PosMix, SizeMix, ShapeBase):
    """Represents an ellipse shape, "draw:ellipse".

    This shape defines an elliptical or circular area.
    """

    _tag = "draw:ellipse"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("cx", "svg:cx"),
        PropDef("cy", "svg:cy"),
        PropDef("rx", "svg:rx"),
        PropDef("ry", "svg:ry"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        kind: str | None = None,
        start_angle: str | None = None,
        end_angle: str | None = None,
        cx: str | None = None,
        cy: str | None = None,
        rx: str | None = None,
        ry: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create an ellipse shape "draw:ellipse".

        Args:
            name: Name of the graphical element.
            style: The style name for the ellipse.
            text_style: The text style name for the ellipse.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the ellipse.
            position: The (x, y) coordinates for the ellipse's position.
            size: The (width, height) for the ellipse's size.
            kind: The appearance of a circle or ellipse, "full", "section",
                "cut" or "arc". Default is "full".
            start_angle: The start angle of a section, cut, or arc where the
                draw:kind is "section", "cut" or "arc".
            end_angle: The end angle of a section, cut, or arc where the
                draw:kind is "section", "cut" or "arc".
            cx: The x-axis coordinate of the center of a circular image map
                area.
            cy: The y-axis coordinate of the center of a circular image map
                area.
            rx: The x-axis radius of the ellipse.
            ry: The y-axis radius of the ellipse.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if position:
                self.position = position
            if size:
                self.size = size
            if kind:
                self.kind = kind
            if start_angle:
                self.start_angle = start_angle
            if end_angle:
                self.end_angle = end_angle
            if cx:
                self.cx = cx
            if cy:
                self.cy = cy
            if rx:
                self.rx = rx
            if ry:
                self.ry = ry


EllipseShape._define_attribut_property()


class CircleShape(AngleMix, PosMix, SizeMix, ShapeBase):
    """Represents an circular shape, "draw:circle".

    This shape defines a circular area.
    """

    _tag = "draw:circle"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("radius", "svg:r"),
        PropDef("cx", "svg:cx"),
        PropDef("cy", "svg:cy"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        radius: str | None = None,
        kind: str | None = None,
        start_angle: str | None = None,
        end_angle: str | None = None,
        center: tuple[str, str] | list[str] | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a circular shape "draw:circle".

        Args:
            name: Name of the graphical element.
            style: The style name for the circle.
            text_style: The text style name for the circle.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the circle.
            position: The (x, y) coordinates for the circle's position.
            size: The (width, height) for the circle's size.
            radius: The radius of the circle.
            kind: The appearance of a circle or circle, "full", "section",
                "cut" or "arc". Default is "full".
            start_angle: The start angle of a section, cut, or arc where the
                draw:kind is "section", "cut" or "arc".
            end_angle: The end angle of a section, cut, or arc where the
                draw:kind is "section", "cut" or "arc".
            center: The x-axis coordinate of the center of a circular image map
                area.
            center: The (cx, cy) coordinates of the center of the circle.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if position:
                self.position = position
            if size:
                self.size = size
            if radius:
                self.radius = radius
            if kind:
                self.kind = kind
            if start_angle:
                self.start_angle = start_angle
            if end_angle:
                self.end_angle = end_angle
            if center:
                self.center = center

    @property
    def center(self) -> tuple[str | None, str | None]:
        "Get or set the center (cx, cy) coordinates of the circle."
        return (self.cx, self.cy)

    @center.setter
    def center(self, center: tuple[str, str] | list[str] | None) -> None:
        if center is None:
            self.cx = None
            self.cy = None
        else:
            self.cx = center[0]
            self.cy = center[1]


CircleShape._define_attribut_property()


class ConnectorShape(ShapeBase):
    """Represents a connector shape, "draw:connector".

    A connector typically links two other shapes.
    """

    _tag = "draw:connector"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("start_shape", "draw:start-shape"),
        PropDef("end_shape", "draw:end-shape"),
        PropDef("start_glue_point", "draw:start-glue-point"),
        PropDef("end_glue_point", "draw:end-glue-point"),
        PropDef("x1", "svg:x1"),
        PropDef("y1", "svg:y1"),
        PropDef("x2", "svg:x2"),
        PropDef("y2", "svg:y2"),
        PropDef("line_skew", "draw:line-skew"),
        PropDef("svg_d", "svg:d"),
        PropDef("view_box", "svg:viewBox"),
    )

    DRAW_TYOE_CHOICE = {  # noqa: RUF012
        "standard",
        "lines",
        "line",
        "curve",
    }

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        connected_shapes: tuple[ShapeBase, ShapeBase] | list[ShapeBase] | None = None,
        glue_points: tuple[str | int, str | int] | list[str | int] | None = None,
        p1: tuple[str, str] | list[str] | None = None,
        p2: tuple[str, str] | list[str] | None = None,
        draw_type: str | None = None,
        line_skew: str | None = None,
        svg_d: str | None = None,
        view_box: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a connector shape "draw:connector".

        Args:
            name: Name of the graphical element.
            style: The style name for the connector.
            text_style: The text style name for the connector.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the connector.
            connected_shapes: A tuple of (start_shape, end_shape)
                where each shape is a shape with a `draw_id`.
            glue_points: A tuple of (start_glue_point, end_glue_point)
                specifying the glue points for connection.
            p1: The (x1, y1) coordinates of the starting point.
            p2: The (x2, y2) coordinates of the ending point.
            draw_type: The line or series of lines that connect two glue
                points. Values are 'standard' (default), 'lines', 'line'
                or 'curve'.
            line_skew: A list of offsets for the placements of connector
                lines if the connector is of type standard.
            svg_d: A path data.
            view_box: The rectangle in a local coordinates system used by the
                points.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if connected_shapes:
                self.connected_shapes = connected_shapes
            if glue_points:
                self.glue_points = glue_points
            self.p1 = p1
            self.p2 = p2
            if draw_type:
                self.draw_type = draw_type
            if line_skew:
                self.line_skew = line_skew
            if svg_d:
                self.svg_d = svg_d
            if view_box:
                self.view_box = view_box

    @property
    def connected_shapes(self) -> tuple[str | None, str | None]:
        """Get or set the connected shapes ("draw:start-shape",
        "draw:end-shape")."""
        get_attr = self.get_attribute_string
        return get_attr("draw:start-shape"), get_attr("draw:end-shape")

    @connected_shapes.setter
    def connected_shapes(
        self, connected_shapes: tuple[ShapeBase, ShapeBase] | list[ShapeBase] | None
    ) -> None:
        if connected_shapes is None:
            self.start_shape = None
            self.end_shape = None
        else:
            self.start_shape = connected_shapes[0].draw_id
            self.end_shape = connected_shapes[1].draw_id

    @property
    def glue_points(self) -> tuple[str | None, str | None]:
        """Get or set the the glue points for connection
        ("draw:start-glue-point", "draw:end-glue-point")."""
        get_attr = self.get_attribute_string
        return get_attr("draw:start-glue-point"), get_attr("draw:end-glue-point")

    @glue_points.setter
    def glue_points(
        self, glue_points: tuple[str | int, str | int] | list[str | int] | None
    ) -> None:
        if glue_points is None:
            self.start_glue_point = None
            self.end_glue_point = None
        else:
            self.start_glue_point = glue_points[0]
            self.end_glue_point = glue_points[1]

    @property
    def p1(self) -> tuple[str | None, str | None]:
        "Get or set the (x1, y1) coordinates of the starting point."
        return (self.x1, self.y1)

    @p1.setter
    def p1(self, p1: tuple[str, str] | list[str] | None) -> None:
        if p1 is None:
            self.x1 = None
            self.y1 = None
        else:
            self.x1 = p1[0]
            self.y1 = p1[1]

    @property
    def p2(self) -> tuple[str | None, str | None]:
        "Get or set the (x2, y2) coordinates of the ending point."
        return (self.x2, self.y2)

    @p2.setter
    def p2(self, p2: tuple[str, str] | list[str] | None) -> None:
        if p2 is None:
            self.x2 = None
            self.y2 = None
        else:
            self.x2 = p2[0]
            self.y2 = p2[1]

    @property
    def draw_type(self) -> str:
        'Get or set the draw type, "draw:type".'
        return self._get_attribute_str_default("draw:type", "standard")

    @draw_type.setter
    def draw_type(self, draw_type: str) -> None:
        if draw_type not in self.DRAW_TYOE_CHOICE:
            raise TypeError(f"'draw:type' not valid: {draw_type!r}")
        self._set_attribute_str_default("draw:type", draw_type, "standard")


ConnectorShape._define_attribut_property()


class DrawControl(PosMix, SizeMix, ShapeBase):
    """Represents a shape that is linked to a control inside an
    "office:forms" element, "draw:control".
    """

    _tag = "draw:control"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control", "draw:control"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        control: str | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a control shape "draw:control".

        Args:
            name: Name of the graphical element.
            style: The style name for the control.
            text_style: The text style name for the control.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the control.
            control: A control within a form that is linked to the control
                shape by its ID.
            position: The (x, y) coordinates for the control's position.
            size: The (width, height) values for the control's size.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if control:
                self.control = control
            if position:
                self.position = position
            if size:
                self.size = size


DrawControl._define_attribut_property()


class DrawPageThumbnail(PosMix, SizeMix, ShapeBase):
    """Represents a rectangular area that displays the thumbnail of a
    drawing page, "draw:page-thumbnail".
    """

    _tag = "draw:page-thumbnail"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDefBool("placeholder", "presentation:placeholder", False),
        PropDef("user_transformed", "presentation:user-transformed"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        text_style: str | None = None,
        draw_id: str | None = None,
        layer: str | None = None,
        page_number: int | None = None,
        position: tuple[str, str] | list[str] | None = None,
        size: tuple[str, str] | list[str] | None = None,
        placeholder: bool | None = None,
        user_transformed: bool | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        transform: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a control shape "draw:page-thumbnail".

        Args:
            name: Name of the graphical element.
            style: The style name for the control.
            text_style: The text style name for the control.
            draw_id: The unique ID for the drawing shape.
            layer: The drawing layer of the control.
            page_number: The number of a page that is displayed as a
                thumbnail.
            position: The (x, y) coordinates for the control's position.
            size: The (width, height) values for the control's size.
            placeholder: True if shape is a placeholder.
            user_transformed: Whether the size and position of the shape
                is set by the user or is set by the corresponding
                presentation shape on the master page.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            transform: White-space or comma separated list of transform
                definitions.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        kwargs.update(
            {
                "name": name,
                "style": style,
                "text_style": text_style,
                "draw_id": draw_id,
                "layer": layer,
                "presentation_class": presentation_class,
                "presentation_style": presentation_style,
                "caption_id": caption_id,
                "class_names": class_names,
                "transform": transform,
                "z_index": z_index,
                "end_cell_address": end_cell_address,
                "end_x": end_x,
                "end_y": end_y,
                "table_background": table_background,
                "anchor_type": anchor_type,
                "anchor_page": anchor_page,
                "xml_id": xml_id,
            }
        )
        super().__init__(**kwargs)
        if self._do_init:
            if page_number:
                self.page_number = page_number
            if position:
                self.position = position
            if size:
                self.size = size
            if placeholder is not None:
                self.placeholder = placeholder
            if user_transformed is not None:
                self.user_transformed = user_transformed

    @property
    def page_number(self) -> int | None:
        """Get or set the number of a page that is displayed as a thumbnail.

        type : int or None
        """
        page_number = self.get_attribute("draw:page-number")
        if page_number is None:
            return None
        return int(page_number)

    @page_number.setter
    def page_number(self, page_number: int | None) -> None:
        self._set_attribute_int("draw:page-number", page_number)


DrawPageThumbnail._define_attribut_property()


class DrawGroup(SvgMixin, AnchorMix, ZMix, Element):
    """Representation of a group of drawing shapes, "draw:g".

    Warning: implementation is currently minimal.

    Drawing shapes contained by a "draw:g" element that is itself
    contained by a "draw:a" element, act as hyperlinks using the
    xlink:href attribute of the containing "draw:a" element. If the
    included drawing shapes are themselves contained within "draw:a"
    elements, then the xlink:href attributes of those "draw:a" elements
    act as the hyperlink information for the shapes they contain.
    """

    _tag = "draw:g"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "draw:name"),
        PropDef("style", "draw:style-name"),
        PropDef("draw_id", "draw:id"),
        PropDef("svg_y", "svg:y"),
        PropDef("presentation_class", "presentation:class-names"),
        PropDef("presentation_style", "presentation:style-name"),
        PropDef("caption_id", "draw:caption-id"),
        PropDef("class_names", "draw:class-names"),
        PropDef("end_cell_address", "table:end-cell-address"),
        PropDef("end_x", "table:end-x"),
        PropDef("end_y", "table:end-y"),
        PropDefBool("table_background", "table:table-background", False),
        PropDef("xml_id", "xml:id"),
    )

    def __init__(
        self,
        name: str | None = None,
        style: str | None = None,
        draw_id: str | None = None,
        svg_y: str | None = None,
        presentation_class: str | None = None,
        presentation_style: str | None = None,
        caption_id: str | None = None,
        class_names: str | None = None,
        z_index: int | None = None,
        end_cell_address: str | None = None,
        end_x: str | None = None,
        end_y: str | None = None,
        table_background: bool | None = None,
        anchor_type: str | None = None,
        anchor_page: int | None = None,
        xml_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a group of drawing shapes "draw:g".

        Args:
            name: Name of the group.
            style: The style name for the group.
            draw_id: The unique ID for the drawing group.
            svg_y: Position on the y-axis of the group.
            presentation_class: White-space-separated list of presentation
                class names.
            presentation_style: Style for a presentation shape.
            caption_id: Target ID assigned to the "draw:text-box" that
                contains the caption.
            class_names: White-space-separated list of styles with the
                family value of graphic.
            z_index: Rendering order for shapes in a document instance.
            end_cell_address: End position of the shape if it is included
                in a spreadsheet document.
            end_x: The x-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            end_y: The y-coordinate of the end position of a shape relative
                to the top-left edge of a cell.
            table_background: Wether the shape is in the table background if
                the drawing shape is included in a spreadsheet.
            anchor_type: How a drawing shape is bound to a text document.
            anchor_page: Physical page number of an anchor if the drawing
                object is bound to a page within a text document.
            xml_id: The unique XML ID.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            if style:
                self.style = style
            if draw_id:
                self.draw_id = draw_id
            if svg_y:
                self.svg_y = svg_y
            if presentation_class:
                self.presentation_class = presentation_class
            if presentation_style:
                self.presentation_style = presentation_style
            if caption_id:
                self.caption_id = caption_id
            if class_names:
                self.class_names = class_names
            if z_index is not None:
                self.z_index = z_index
            if end_cell_address:
                self.end_cell_address = end_cell_address
            if end_x:
                self.end_x = end_x
            if end_y:
                self.end_y = end_y
            if table_background is not None:
                self.table_background = table_background
            if anchor_type:
                self.anchor_type = anchor_type
            if anchor_page is not None:
                self.anchor_page = anchor_page
            if xml_id:
                self.xml_id = xml_id

    def get_formatted_text(self, context: dict | None = None) -> str:
        """Return the formatted text content of the shape.

        Args:
            context: A dictionary of context variables.

        Returns:
            str: The formatted text content.
        """
        result: list[str] = []
        for child in self.children:
            result.append(child.get_formatted_text(context))
        result.append("\n")
        return "".join(result)


DrawGroup._define_attribut_property()

registered_shapes = [
    s._tag
    for s in (
        CircleShape,
        ConnectorShape,
        DrawCaption,
        DrawControl,
        DrawMeasure,
        DrawPageThumbnail,
        DrawPath,
        EllipseShape,
        LineShape,
        PolygonShape,
        PolylineShape,
        RectangleShape,
        RegularPolygonShape,
    )
]
register_element_class(CircleShape)
register_element_class(ConnectorShape)
register_element_class(DrawCaption)
register_element_class(DrawControl)
register_element_class(DrawGroup)
register_element_class(DrawMeasure)
register_element_class(DrawPageThumbnail)
register_element_class(DrawPath)
register_element_class(EllipseShape)
register_element_class(LineShape)
register_element_class(PolygonShape)
register_element_class(PolylineShape)
register_element_class(RectangleShape)
register_element_class(RegularPolygonShape)



#!/usr/bin/env python
"""Build a commercial document, with numerical values displayed in
both the text and in a table.
"""

import os
from pathlib import Path

from odfdo import (
    Cell,
    Document,
    Header,
    List,
    ListItem,
    Paragraph,
    Row,
    Table,
    create_table_cell_style,
    make_table_cell_border_string,
)

_DOC_SEQUENCE = 50
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "commercial"
TARGET = "commercial.odt"
TAX_RATE = 0.20


class Product:
    """Minimalistic Product."""

    def __init__(self, reference: int, name: str, price: float) -> None:
        self.reference = reference
        self.name = f"Product {name}"
        self.price = price


class OrderLine:
    """Line of an Order."""

    def __init__(self, reference: int, quantity: int) -> None:
        self.reference = reference
        self.quantity = quantity


def make_product_catalog() -> list[Product]:
    """Generate a list of Product."""
    catalog: list[Product] = []
    price = 10.0
    for index in range(5):
        catalog.append(Product(index, chr(65 + index), price))
        price += 10.5
    return catalog


def make_order(catalog: list[Product]) -> list[OrderLine]:
    """Generate purchase order list."""
    order: list[OrderLine] = []
    quantity = 1
    for product in catalog:
        quantity = int(quantity * 2.5)
        order.append(OrderLine(product.reference, quantity))
    return order


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def create_header_cell_style(doc: Document) -> str:
    """Create a Cell style, insert it in Document, return its name."""
    border = make_table_cell_border_string(thick="0.03cm", color="black")
    cell_style = create_table_cell_style(
        color="black",
        background_color=(210, 210, 210),
        border_right=border,
        border_left=border,
        border_bottom=border,
        border_top=border,
    )
    style_name = doc.insert_style(style=cell_style, automatic=True)
    return style_name


def add_top_content(doc: Document, catalog: list[Product]) -> None:
    """Add some descriptive content to the document."""
    body = doc.body

    title1 = Header(1, "Basic commercial document")
    body.append(title1)
    title11 = Header(2, "Available products")
    body.append(title11)
    paragraph = Paragraph("Here the list:")
    body.append(paragraph)
    # List of products in a list :
    product_list = List()  # odfdo.List
    body.append(product_list)
    for product in catalog:
        item = ListItem(f"{product.name:<10}, price: {product.price:.2f} €")
        product_list.append(item)


def add_order_table(
    doc: Document, catalog: list[Product], order: list[OrderLine]
) -> None:
    """Add a table with order lines."""
    body = doc.body
    title12 = Header(2, "Your order")
    body.append(title12)

    style_name = create_header_cell_style(doc)
    table = make_order_table(catalog, order, style_name)
    body.append(table)


def make_order_table(
    catalog: list[Product],
    order: list[OrderLine],
    style_name: str,
) -> Table:
    """Build the order table."""
    table = Table("Table")

    # Header of table
    row = Row()
    row.set_values(["Product", "Price", "Quantity", "Amount"])
    table.set_row("A1", row)
    # or: table.set_row(0, row)

    # Add a row for each order line
    row_number = 0
    for line in order:
        row_number += 1
        product = catalog[line.reference]

        row = Row()

        row.set_value("A", product.name)
        # or : row.set_value(0, product.name)

        cell = Cell()
        cell.set_value(
            product.price,
            text=f"{product.price:.2f} €",
            currency="EUR",
            cell_type="float",
        )
        row.set_cell("B", cell)
        # or : row.set_cell(1, cell)

        row.set_value("C", line.quantity)
        # row.set_value(2, line.quantity)

        price = product.price * line.quantity
        cell = Cell()
        cell.set_value(
            price,
            text=f"{price:.2f} €",
            currency="EUR",
            cell_type="float",
        )
        row.set_cell("D", cell)

        table.set_row(row_number, row)

    # Total lines

    # add a merged empty row
    row = Row()
    row_number += 1
    table.set_row(row_number, row)
    table.set_span((0, row_number, 3, row_number))

    # compute total line
    row = Row()
    row_number += 1
    row.set_value(0, "Total:")
    total = sum(table.get_column_values(3)[1:-1])
    # note: total is a Decimal
    cell = Cell()
    cell.set_value(
        total,
        text=f"{total:.2f} €",
        currency="EUR",
        cell_type="float",
    )
    row.set_cell(3, cell)
    table.set_row(row_number, row)
    # merge the 3 first columns for this row:
    table.set_span((0, row_number, 2, row_number), merge=True)

    # compute VAT line
    row = Row()
    row_number += 1
    row.set_value(0, "Total with tax:")
    total_vat = float(total) * (1 + TAX_RATE)
    cell = Cell()
    cell.set_value(
        total_vat,
        text=f"{total_vat:.2f} €",
        currency="EUR",
        cell_type="float",
    )
    row.set_cell(3, cell)
    table.set_row(row_number, row)
    table.set_span((0, row_number, 2, row_number), merge=True)

    # Let's add some style on header row
    row = table.get_row(0)
    for cell in row.iter_cells():
        cell.style = style_name
        row.set_cell(x=cell.x, cell=cell)
    table.set_row(row.y, row)

    return table


def generate_commercial(catalog: list[Product], order: list[OrderLine]) -> Document:
    """Generate a Text Document with table in in."""
    document = Document("text")
    add_top_content(document, catalog)
    add_order_table(document, catalog, order)

    return document


def main() -> None:
    catalog = make_product_catalog()
    order = make_order(catalog)
    document = generate_commercial(catalog, order)
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    table = document.body.get_table(name="Table")
    assert isinstance(table, Table)
    assert table.get_cell("A1").value == "Product"
    assert table.get_cell("A2").value == "Product A"
    assert table.get_cell("A8").value == "Total:"
    assert table.get_cell("B1").value == "Price"
    assert table.get_cell("C1").value == "Quantity"
    assert table.get_cell("C2").value == 2
    assert table.get_cell("D1").value == "Amount"


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python


from odfdo.meta_hyperlink_behaviour import MetaHyperlinkBehaviour


def test_create_minimal():
    behaviour = MetaHyperlinkBehaviour()
    expected = '<meta:hyperlink-behaviour office:target-frame-name="_blank" xlink:show="replace"/>'
    assert behaviour.serialize() == expected
    assert behaviour.show == "replace"


def test_create():
    behaviour = MetaHyperlinkBehaviour(target_frame_name="some_frame", show="_top")
    expected = '<meta:hyperlink-behaviour office:target-frame-name="some_frame" xlink:show="_top"/>'
    assert behaviour.serialize() == expected
    assert behaviour.show == "_top"
    assert behaviour.target_frame_name == "some_frame"
    assert repr(behaviour) == (
        "<MetaHyperlinkBehaviour tag=meta:hyperlink-behaviour target=some_frame show=_top>"
    )


def test_as_dict():
    behaviour = MetaHyperlinkBehaviour(target_frame_name="some_frame", show="_top")
    expected = {
        "office:target-frame-name": "some_frame",
        "xlink:show": "_top",
    }
    assert behaviour.as_dict() == expected


def test_from_dict():
    behaviour = MetaHyperlinkBehaviour(target_frame_name="initial_frame", show="")
    imported = {
        "office:target-frame-name": "new_frame",
        "xlink:show": "_top",
    }
    behaviour.from_dict(imported)
    assert behaviour.as_dict() == imported


def test_str():
    behaviour = MetaHyperlinkBehaviour(target_frame_name="some_frame", show="_top")
    expected = "(some_frame)"
    assert str(behaviour) == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Config elements for the "settings.xml" part.

This module defines classes representing configuration elements found within
the `settings.xml` file of an ODF document, specifically those related to
`config:config-item-set`.
"""

from __future__ import annotations

from typing import Any, ClassVar, cast

from .datatype import Boolean
from .element import (
    Element,
    PropDef,
    PropDefBool,
    class_from_tag,
    register_element_class,
)


def _as_dict(
    element: ConfigItemMapEntry
    | ConfigItemMapIndexed
    | ConfigItemMapNamed
    | ConfigItemSet,
) -> dict[str, str | int | bool | list[Any] | dict[str, Any]]:
    """Internal helper to serialize a configuration element to a dictionary."""
    conf: dict[str, str | int | bool | list[Any] | dict[str, Any]] = {
        "class": element._tag
    }
    if element.name:
        conf["config:name"] = element.name
    # all children are known to be classes with as_dict()
    children = [child.as_dict() for child in element.children]  # type: ignore[attr-defined]
    if children:
        conf["children"] = children
    return conf


def _from_dict(data: dict[str, str | int | bool | dict[str, Any]]) -> Element:
    """Internal helper to deserialize a dictionary into a configuration
    element.
    """
    class_tag: str = data.pop("class")  # type: ignore[assignment]
    if class_tag == "config:config-item":
        return ConfigItem.from_dict(data)  # type: ignore[arg-type]
    kwargs: dict[str, str | int | bool] = {}
    if "config:name" in data:
        kwargs["name"] = data.pop("config:name")  # type: ignore[assignment]
    children: list[Any] = data.pop("children", [])  # type: ignore[assignment]
    klass = class_from_tag(class_tag)
    result = klass(**kwargs)
    for child in children:
        result.append(_from_dict(child))
    return result


class ConfigItemSet(Element):
    """Represents a container element for application setting elements,
    "config:config-item-set" tag.

    This element can contain other configuration items, item maps, or item
    sets, forming a hierarchical structure for document settings.

    Attributes:
        name (str): The name of the configuration item set.
    """

    _tag: str = "config:config-item-set"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("name", "config:name"),)

    def __init__(
        self,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ConfigItemSet element.

        This element acts as a container for various application setting
        elements, including `config:config-item`,
        `config:config-item-map-indexed`, `config:config-item-map-named`,
        and nested `config:config-item-set`.

        The "config:config-item-set" element has the following child elements:
        "config:config-item", "config:config-item-map-indexed",
        "config:config-item-map-named" and "config:config-item-set"

        Args:
            name: The name of the configuration item set.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.name = name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    @property
    def config_item_sets(self) -> list[ConfigItemSet]:
        """Get the list of nested ConfigItemSet elements.

        Returns:
            list[ConfigItemSet]: A list of `ConfigItemSet` objects.
        """
        return cast(list[ConfigItemSet], self.get_elements("config:config-item-set"))

    @property
    def config_item_maps_indexed(self) -> list[ConfigItemMapIndexed]:
        """Get the list of indexed configuration item maps.

        Returns:
            list[ConfigItemMapIndexed]: A list of `ConfigItemMapIndexed`
                objects.
        """
        return cast(
            list[ConfigItemMapIndexed],
            self.get_elements("config:config-item-map-indexed"),
        )

    @property
    def config_item_maps_named(self) -> list[ConfigItemMapNamed]:
        """Get the list of named configuration item maps.

        Returns:
            list[ConfigItemMapNamed]: A list of `ConfigItemMapNamed` objects.
        """
        return cast(
            list[ConfigItemMapNamed], self.get_elements("config:config-item-map-named")
        )

    @property
    def config_items(self) -> list[ConfigItem]:
        """Get the list of individual configuration items.

        Returns:
            list[ConfigItem]: A list of `ConfigItem` objects.
        """
        return cast(list[ConfigItem], self.get_elements("config:config-item"))

    def as_dict(self) -> dict[str, str | int | bool | list[Any] | dict[str, Any]]:
        """Serialize the element to a dictionary.

        Returns:
            A dict with content of the ConfigItemSet serialized.
        """
        return _as_dict(self)

    @classmethod
    def from_dict(
        cls, data: dict[str, str | int | bool | dict[str, Any]]
    ) -> ConfigItemSet:
        """Create an element from a dictionary.

        Args:
            data: The ConfigItemSet content serialized as a dict.

        Returns:
            A ConfigItemSet.
        """
        return cast(ConfigItemSet, _from_dict(data))


ConfigItemSet._define_attribut_property()


class ConfigItemMapIndexed(Element):
    """Represents a container for ordered sequences of application settings,
    corresponding to the "config:config-item-map-indexed" tag.

    This element is used to store a collection of configuration items that are
    ordered and can be accessed by index.

    Attributes:
        name (str): The name of the indexed configuration item map.
    """

    _tag: str = "config:config-item-map-indexed"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("name", "config:name"),)

    def __init__(
        self,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ConfigItemMapIndexed element.

        This container holds ordered sequences of application settings.

        The "config:config-item-map-indexed" element has the following child
        element: "config:config-item-map-entry".

        Args:
            name: The name of the indexed configuration item map.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.name = name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    @property
    def config_item_maps_entries(self) -> list[ConfigItemMapEntry]:
        """Get the list of configuration item map entries.

        Returns:
            list[ConfigItemMapEntry]: A list of `ConfigItemMapEntry` objects.
        """
        return cast(
            list[ConfigItemMapEntry], self.get_elements("config:config-item-map-entry")
        )

    def as_dict(self) -> dict[str, str | int | bool | list[Any] | dict[str, Any]]:
        """Serialize the element to a dictionary.

        Returns:
            A dict with content of the ConfigItemMapIndexed serialized.
        """
        return _as_dict(self)

    @classmethod
    def from_dict(
        cls, data: dict[str, str | int | bool | dict[str, Any]]
    ) -> ConfigItemMapIndexed:
        """Create an element from a dictionary.

        Args:
            data: The ConfigItemMapIndexed content serialized as a dict.

        Returns:
            A ConfigItemMapIndexed.
        """
        return cast(ConfigItemMapIndexed, _from_dict(data))


ConfigItemMapIndexed._define_attribut_property()


class ConfigItemMapEntry(Element):
    """Represents a single setting entry in a sequence of settings,
    corresponding to the "config:config-item-map-entry" tag.

    The setting itself is defined by the child element of "config:config-item-map-entry",
    and may be a single value, a set of settings, or a sequence of settings.

    The "config:config-item-map-entry" element has the following child elements:
    "config:config-item", "config:config-item-map-indexed", "config:config-item-map-named",
    and "config:config-item-set.

    Attributes:
        name (str): The name of the entry.
    """

    _tag: str = "config:config-item-map-entry"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("name", "config:name"),)

    def __init__(
        self,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ConfigItemMapEntry element.

        This represents an entry within an ordered configuration map.

        Args:
            name: The name of the entry.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.name = name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    @property
    def config_item_sets(self) -> list[ConfigItemSet]:
        """Get the list of nested ConfigItemSet elements.

        Returns:
            list[ConfigItemSet]: A list of `ConfigItemSet` objects.
        """
        return cast(list[ConfigItemSet], self.get_elements("config:config-item-set"))

    @property
    def config_item_maps_indexed(self) -> list[ConfigItemMapIndexed]:
        """Get the list of indexed configuration item maps.

        Returns:
            list[ConfigItemMapIndexed]: A list of `ConfigItemMapIndexed`
                objects.
        """
        return cast(
            list[ConfigItemMapIndexed],
            self.get_elements("config:config-item-map-indexed"),
        )

    @property
    def config_item_maps_named(self) -> list[ConfigItemMapNamed]:
        """Get the list of named configuration item maps.

        Returns:
            list[ConfigItemMapNamed]: A list of `ConfigItemMapNamed` objects.
        """
        return cast(
            list[ConfigItemMapNamed], self.get_elements("config:config-item-map-named")
        )

    @property
    def config_items(self) -> list[ConfigItem]:
        """Get the list of individual configuration items.

        Returns:
            list[ConfigItem]: A list of `ConfigItem` objects.
        """
        return cast(list[ConfigItem], self.get_elements("config:config-item"))

    def as_dict(self) -> dict[str, str | int | bool | list[Any] | dict[str, Any]]:
        """Serialize the element to a dictionary.

        Returns:
            A dict with content of the ConfigItemMapEntry serialized."""
        return _as_dict(self)

    @classmethod
    def from_dict(
        cls, data: dict[str, str | int | bool | dict[str, Any]]
    ) -> ConfigItemMapEntry:
        """Create an element from a dictionary.

        Args:
            data: The ConfigItemMapEntry content serialized as a dict.

        Returns:
            A ConfigItemMapEntry.
        """
        return cast(ConfigItemMapEntry, _from_dict(data))


ConfigItemMapEntry._define_attribut_property()


class ConfigItemMapNamed(Element):
    """Represents a container for a sequence of application setting elements,
    where each sequence is identified by the value of its `config:name`
    attribute. This corresponds to the "config:config-item-map-named" tag.

    This element is used to store a collection of configuration items that
    can be accessed by a descriptive name rather than an index.

    The "config:config-item-map-named" element has the following child
    element: "config:config-item-map-entry".

    Attributes:
        name (str): The name of the named configuration item map.
    """

    _tag: str = "config:config-item-map-named"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("name", "config:name"),)

    def __init__(
        self,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ConfigItemMapNamed element.

        This container holds named sequences of application setting elements.

        Args:
            name: The name of the named configuration item map.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.name = name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    @property
    def config_item_maps_entries(self) -> list[ConfigItemMapEntry]:
        """Get the list of configuration item map entries.

        Returns:
            list[ConfigItemMapEntry]: A list of `ConfigItemMapEntry` objects.
        """
        return cast(
            list[ConfigItemMapEntry], self.get_elements("config:config-item-map-entry")
        )

    def as_dict(self) -> dict[str, str | int | bool | list[Any] | dict[str, Any]]:
        """Serialize the element to a dictionary.

        Returns:
            A dict with content of the ConfigItemMapNamed serialized.
        """
        return _as_dict(self)

    @classmethod
    def from_dict(
        cls, data: dict[str, str | int | bool | dict[str, Any]]
    ) -> ConfigItemMapNamed:
        """Create an element from a dictionary.

        Args:
            data: The ConfigItemMapNamed content serialized as a dict.

        Returns:
            A ConfigItemMapNamed.
        """
        return cast(ConfigItemMapNamed, _from_dict(data))


ConfigItemMapNamed._define_attribut_property()


class ConfigItem(Element):
    """Represents an element containing the value of an application setting,
    identified by its `config:name` attribute. This corresponds to the
    "config:config-item" tag.

    This element holds a single configuration value and does not have any
    child elements.

    Attributes:
        name (str): The name of the configuration item.
        config_type (str): The data type of the configuration item's value.
        value (str | int | bool): The actual value of the configuration item.
    """

    _tag: str = "config:config-item"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "config:name"),
        PropDef("config_type", "config:type"),
    )
    _properties = (PropDef("name", "config:name"),)
    TYPES: ClassVar = {
        "boolean",
        "short",
        "int",
        "long",
        "double",
        "string",
        "datetime",
        "base64Binary",
    }

    def __init__(
        self,
        name: str | None = None,
        config_type: str | None = None,
        value: str | int | bool | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a ConfigItem element.

        This element contains the value of an application setting.

        Args:
            name: The name of the configuration item.
            config_type: The data type of the configuration item's value,
                one of "boolean", "short", "int", "long", "double", "string",
                "datetime", or "base64Binary".
            value: The actual value of the configuration item.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.name = name
            self.config_type = config_type
            self.value = value

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    @property
    def config_type(self) -> str:
        """Get the data type of the configuration item's value."""
        config_type = self.get_attribute_string("config:type")
        if config_type not in self.TYPES:
            return "string"
        return config_type

    @config_type.setter
    def config_type(self, config_type: str | None) -> None:
        """Set the data type of the configuration item's value."""
        if config_type not in self.TYPES:
            config_type = "string"
        self._set_attribute_str("config:type", config_type)

    @property
    def value(self) -> str | int | bool:
        """Get or set the value of the configuration item.

        When getting, the value is cast to its appropriate Python type (str,
        int, or bool) based on the `config_type` attribute.

        When setting, the provided value is converted to a string and stored
        as the element's text content, applying type-specific encoding (e.g.,
        for booleans) based on the `config_type` attribute.

        Returns:
            The value of the configuration item as a str, int, or bool.
        """
        content: str = self.text
        config_type = self.config_type
        if config_type == "boolean":
            return Boolean.decode(content)
        elif config_type in {"short", "int", "long", "double"}:
            return int(content)
        return content or ""

    @value.setter
    def value(self, value: str | int | bool | None) -> None:
        config_type = self.config_type
        if config_type == "boolean":
            self.text = Boolean.encode(value)
        elif config_type in {"short", "int", "long", "double"}:
            self.text = str(int(value))  # type: ignore[arg-type]
        else:
            self.text = str(value or "")

    def as_dict(self) -> dict[str, str | int | bool]:
        """Serialize the element to a dictionary.

        Returns:
            A dict with content of the ConfigItem serialized.
        """
        return {
            "class": self._tag,
            "config:name": self.name,  # type: ignore[dict-item]
            "config:type": self.config_type,
            "value": self.value,
        }

    @classmethod
    def from_dict(cls, data: dict[str, str | int | bool]) -> ConfigItem:
        """Create an element from a dictionary.

        Args:
            data: The  ConfigItem content serialized in a dict.

        Returns:
            A ConfigItem instance."
        """
        return cls(
            name=data["config:name"],  # type: ignore[arg-type]
            config_type=data.get("config:type"),  # type: ignore[arg-type]
            value=data.get("value"),
        )


ConfigItem._define_attribut_property()

register_element_class(ConfigItem)
register_element_class(ConfigItemMapEntry)
register_element_class(ConfigItemMapIndexed)
register_element_class(ConfigItemMapNamed)
register_element_class(ConfigItemSet)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
"""Classes for ODF implementation of SMIL (Synchronized Multimedia Integration
Language).
"""

from __future__ import annotations

from typing import Any

from .element import Element, PropDef, register_element_class


class AnimPar(Element):
    """Container for SMIL animations, "anim:par"."""

    _tag = "anim:par"
    _properties = (
        PropDef("presentation_node_type", "presentation:node-type"),
        PropDef("smil_begin", "smil:begin"),
    )

    def __init__(
        self,
        presentation_node_type: str | None = None,
        smil_begin: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a container for SMIL presentation animations "anim:par".

        Args:
            presentation_node_type: The type of presentation node.
                Can be 'default', 'on-click', 'with-previous', 'after-previous',
                'timing-root', 'main-sequence', or 'interactive-sequence'.
            smil_begin: Specifies the begin time of the animation.
                Can be 'indefinite', a time value (e.g., '10s'), or an ID reference
                (e.g., '[id].click', '[id].begin').
        """
        super().__init__(**kwargs)
        if self._do_init:
            if presentation_node_type:
                self.presentation_node_type = presentation_node_type
            if smil_begin:
                self.smil_begin = smil_begin


AnimPar._define_attribut_property()


class AnimSeq(Element):
    """Container for SMIL Presentation Animations, "anim:seq".

    Animations inside this block are executed after the slide has executed its
    initial transition.
    """

    _tag = "anim:seq"
    _properties = (PropDef("presentation_node_type", "presentation:node-type"),)

    def __init__(
        self,
        presentation_node_type: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a container for SMIL Presentation Animations "anim:seq".

        Animations inside this block are executed after the slide
        has executed its initial transition.

        Args:
            presentation_node_type: The type of presentation node.
                Can be 'default', 'on-click', 'with-previous', 'after-previous',
                'timing-root', 'main-sequence', or 'interactive-sequence'.
        """
        super().__init__(**kwargs)
        if self._do_init and presentation_node_type:
            self.presentation_node_type = presentation_node_type


AnimSeq._define_attribut_property()


class AnimTransFilter(Element):
    """Transition between two frames, "anim:transitionFilter"."""

    _tag = "anim:transitionFilter"
    _properties = (
        PropDef("smil_dur", "smil:dur"),
        PropDef("smil_type", "smil:type"),
        PropDef("smil_subtype", "smil:subtype"),
        PropDef("smil_direction", "smil:direction"),
        PropDef("smil_fadeColor", "smil:fadeColor"),
        PropDef("smil_mode", "smil:mode"),
    )

    def __init__(
        self,
        smil_dur: str | None = None,
        smil_type: str | None = None,
        smil_subtype: str | None = None,
        smil_direction: str | None = None,
        smil_fadeColor: str | None = None,
        smil_mode: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a transition between two frames "anim:transitionFilter".

        Args:
            smil_dur: The duration of the transition.
            smil_type: The type of transition. See
                http://www.w3.org/TR/SMIL20/smil-transitions.html#TransitionEffects-Appendix
                for a list of types.
            smil_subtype: The subtype of transition. See
                http://www.w3.org/TR/SMIL20/smil-transitions.html#TransitionEffects-Appendix
                for a list of subtypes.
            smil_direction: The direction of the transition. Can be
                'forward' or 'reverse'.
            smil_fadeColor: The fade color for the transition.
            smil_mode: The mode of the transition. Can be 'in' or 'out'.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if smil_dur:
                self.smil_dur = smil_dur
            if smil_type:
                self.smil_type = smil_type
            if smil_subtype:
                self.smil_subtype = smil_subtype
            if smil_direction:
                self.smil_direction = smil_direction
            if smil_fadeColor:
                self.smil_fadeColor = smil_fadeColor
            if smil_mode:
                self.smil_mode = smil_mode


AnimTransFilter._define_attribut_property()

register_element_class(AnimPar)
register_element_class(AnimSeq)
register_element_class(AnimTransFilter)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Mixin class for Style properties methods."""

from __future__ import annotations

import contextlib
from typing import Any, ClassVar
from warnings import warn

from .element import Element
from .style_base import PropDict, StyleBase
from .style_utils import _expand_properties_dict, _expand_properties_list, _merge_dicts
from .utils.style_constants import STYLE_ATTRIBUTES


class StyleProps(StyleBase):
    """Mixin for Style properties methods.

    This class provides methods for getting, setting, and deleting properties
    associated with various style areas within an ODF document. It leverages
    `StyleBase` for fundamental style functionalities.
    """

    AREAS: ClassVar[set[str]] = {
        "chart",
        "drawing-page",
        "graphic",
        "header-footer",
        "list-level",
        "page-layout",
        "paragraph",
        "ruby",
        "section",
        "table",
        "table-cell",
        "table-column",
        "table-row",
        "text",
    }

    def _check_area(self, area: str | None) -> str:
        """Validate the provided area or use the style's family as default.

        Args:
            area: The area to check. If None, uses the style's family.

        Returns:
            str: The validated area.

        Raises:
            ValueError: If the area is not a recognized type.
        """
        if area is None:
            area = self.family
        if area not in StyleProps.AREAS:
            raise ValueError(f"Unexpected area value: {area!r}")
        return area

    def get_properties(self, area: str | None = None) -> PropDict | None:
        """Get the mapping of all properties of this style.

        By default, retrieves properties of the same family as the style (e.g.,
        paragraph properties for a paragraph style). Specify the `area` to get
        properties from a different area (e.g., text properties of a paragraph style).

        Args:
            area: The specific area of properties to retrieve
                (e.g., 'text', 'paragraph').

        Returns:
            dict[str, str | dict] | None: A dictionary of properties, or None if no properties are found.
        """
        try:
            area = self._check_area(area)
        except ValueError:
            return None
        element = self.get_element(f"style:{area}-properties")
        if element is None:
            return None
        properties: PropDict = dict(element.attributes)
        # Nested properties are nested dictionaries
        for child in element.children:
            properties[child.tag] = dict(child.attributes)
        return properties

    @staticmethod
    def _update_boolean_styles(props: dict[str, str | bool]) -> None:
        """Update a dictionary of style properties with boolean values for common text attributes.

        This static method adds or updates 'color', 'background_color', 'italic',
        'bold', 'fixed', 'underline', and 'strike' keys in the provided `props`
        dictionary based on existing OpenDocument style attributes.

        Args:
            props: The dictionary of style properties to update.
        """
        strike = props.get("style:text-line-through-style", "")
        if strike == "none":
            strike = ""
        underline = props.get("style:text-underline-style", "")
        if underline == "none":
            underline = ""
        props.update(
            {
                "color": props.get("fo:color") or "",
                "background_color": props.get("fo:background-color") or "",
                "italic": props.get("fo:font-style", "") == "italic",
                "bold": props.get("fo:font-weight", "") == "bold",
                "fixed": props.get("style:font-pitch", "") == "fixed",
                "underline": bool(underline),
                "strike": bool(strike),
            }
        )

    def get_list_style_properties(self) -> dict[str, str | bool]:
        """Get list style properties as a dictionary with enhanced values.

        Enhanced values returned:
        - "color": str
        - "background_color": str
        - "italic": bool
        - "bold": bool
        - "fixed": bool
        - "underline": bool
        - "strike": bool

        Returns:
            dict[str, str | bool]: A dictionary containing list style properties.
        """
        return self.get_text_properties()

    def get_text_properties(self) -> dict[str, str | bool]:
        """Get text properties of style as a dictionary with enhanced values.

        Enhanced values returned:
        - "color": str
        - "background_color": str
        - "italic": bool
        - "bold": bool
        - "fixed": bool
        - "underline": bool
        - "strike": bool

        Returns:
            dict[str, str | bool]: A dictionary containing text properties.
        """
        props = self.get_properties(area="text") or {}
        self._update_boolean_styles(props)  # type: ignore[arg-type]
        return props  # type: ignore[return-value]

    @staticmethod
    def _apply_valid_properties(
        properties_element: Element,
        area: str,
        properties: PropDict,
    ) -> None:
        # first filter only valid known properties
        allowed = STYLE_ATTRIBUTES.get(area)
        if allowed:
            for key, value in properties.items():
                if key in allowed:
                    if isinstance(value, (str, bool, tuple)):
                        properties_element.set_attribute(key, value)
                else:
                    msg = f"{key!r} property not allowed in <{properties_element.tag}>"
                    warn(msg, stacklevel=2)

        else:
            for key, value in properties.items():
                if isinstance(value, (str, bool, tuple)):
                    properties_element.set_attribute(key, value)

    def set_properties(
        self,
        properties: PropDict | None = None,
        style: StyleBase | None = None,
        area: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Set the properties of the "area" type of this style.

        Properties are given either as a dict or as named arguments (or both).
        The area is identical to the style family by default. If the
        properties element is missing, it is created.

        Instead of properties, you can pass a style with properties of the
        same area. These will be copied.

        Args:
            properties (dict, optional): A dictionary of properties to set.
            style (StyleBase, optional): Another StyleBase object from which
                to copy properties.
            area (str, optional): The specific area of properties to set
                (e.g., 'paragraph', 'text').
            **kwargs: Arbitrary keyword arguments representing properties to set.
        """
        area = self._check_area(area)
        if properties is None:
            properties = {}
        element = self.get_element(f"style:{area}-properties")
        if element is None:
            element = Element.from_tag(f"style:{area}-properties")
            self.append(element)
        working_properties: PropDict
        if properties or kwargs:
            working_properties = _expand_properties_dict(
                _merge_dicts(properties, kwargs)
            )
        elif style is not None:
            style_props = style.get_properties(area=area)
            if style_props is None:
                return
            working_properties = style_props
        else:
            working_properties = {}
        for key, value in working_properties.items():
            if value is None:
                with contextlib.suppress(KeyError):
                    element.del_attribute(key)
        self._apply_valid_properties(element, area, working_properties)

    def del_properties(
        self,
        properties: list[str] | None = None,
        area: str | None = None,
    ) -> None:
        """Delete the given properties from the specified `area`.

        Properties can be specified either as a list argument or as positional
        arguments. Properties are removed only from the given area, which
        defaults to the style's family.

        Args:
            properties: A list of property names to delete.
            area: The specific area from which to delete properties.
        """
        area = self._check_area(area)
        if properties is None:
            properties = []
        element = self.get_element(f"style:{area}-properties")
        if element is None:
            raise ValueError(
                f"The Properties element is non-existent for: style:{area}-properties"
            )
        for key in _expand_properties_list(properties):
            with contextlib.suppress(KeyError):
                element.del_attribute(key)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Tests for flat XML conversion with missing parts.

This tests that flat ODF files with missing parts (like settings.xml,
styles.xml) can be converted to standard ODF format.
"""

from odfdo import Document


class TestFlatXmlWithMissingParts:
    """Test flat XML conversion when some parts are missing."""

    def test_flat_xml_with_document_meta_wrapper(self, tmp_path, samples):
        """Flat XML with office:document-meta wrapper should be handled."""
        xml_path = samples("example.xml")
        doc = Document(xml_path)

        # Should have meta.xml extracted
        parts = doc.container.get_parts()
        assert "meta.xml" in parts

    def test_flat_xml_to_odt_with_missing_parts(self, tmp_path, samples):
        """Converting flat XML with missing parts to ODT should work."""
        xml_path = samples("example.xml")
        doc = Document(xml_path)

        odt_path = tmp_path / "example.odt"
        doc.save(odt_path, packaging="zip")

        loaded = Document(odt_path)
        assert loaded.container.mimetype == "application/vnd.oasis.opendocument.text"

    def test_flat_xml_missing_settings_styles(self, tmp_path, samples):
        """Flat XML without settings.xml and styles.xml should convert."""
        xml_path = samples("example.xml")
        doc = Document(xml_path)

        # Check that settings and styles are not in parts
        parts = doc.container.get_parts()
        assert "settings.xml" not in parts
        assert "styles.xml" not in parts

        # Should still be able to save
        odt_path = tmp_path / "example.odt"
        doc.save(odt_path, packaging="zip")

        # Verify the saved ODT
        loaded = Document(odt_path)
        saved_parts = loaded.container.get_parts()
        assert "content.xml" in saved_parts
        assert "meta.xml" in saved_parts



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""Representation of the "content.xml" part."""

from __future__ import annotations

from typing import TYPE_CHECKING

from odfdo.element import Element

from .xmlpart import XmlPart

if TYPE_CHECKING:
    from .style_base import StyleBase


class Content(XmlPart):
    """Representation of the "content.xml" part."""

    def _get_style_contexts(self, family: str | None) -> tuple:
        if family == "font-face":
            return (self.get_element("//office:font-face-decls"),)
        return (
            self.get_element("//office:font-face-decls"),
            self.get_element("//office:automatic-styles"),
        )

    def __str__(self) -> str:
        return str(self.body)

    # Public API

    def get_styles(self, family: str | None = None) -> list[StyleBase]:
        """Return the list of styles in the Content part, optionally limited to
        the given family.

        Args:
            family: The style family to filter by.

        Returns:
            A list of styles.
        """
        result: list[StyleBase] = []
        for context in self._get_style_contexts(family):
            if context is None:
                continue
            result.extend(context.get_styles(family=family))
        return result

    def get_style(
        self,
        family: str,
        name_or_element: str | Element | None = None,
        display_name: str | None = None,
    ) -> StyleBase | None:
        """Return the style uniquely identified by the name/family pair. If the
        argument is already a style object, it will return it.

        If the name is None, the default style is fetched.

        If the name is not the internal name but the name you gave in the
        desktop application, use display_name instead.

        Args:
            family: The family of the style to retrieve (e.g., 'paragraph',
                'text', 'graphic').
            name_or_element: The name of the style or a Style element. If
                None, the default style for the family is returned.
            display_name: The display name of the style to search for.

        Returns:
            The Style object, or None if not found.
        """
        for context in self._get_style_contexts(family):
            if context is None:
                continue
            style: StyleBase | None = context.get_style(
                family,
                name_or_element=name_or_element,
                display_name=display_name,
            )
            if style is not None:
                return style
        return None



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Forms class for "form:form" tag."""

from __future__ import annotations

from typing import Any, cast

from .element import Element, PropDef, register_element_class
from .form_controls_mixins import OfficeTargetFrameMixin


class FormMixin(Element):
    """Mixin class for classes containing Form.

    Used by the following classes: "form:form", "office:forms".
    And also in "office:text", "table:table" for ease of use.
    """

    def get_forms(
        self,
        name: str | None = None,
    ) -> list[Form]:
        """Return all the Form that match the criteria.

        Args:
            name: The name of the form to filter by.

        Returns:
            list[Form]: A list of Form instances matching the criteria.
        """
        return cast(
            list[Form],
            self._filtered_elements("descendant::form:form", form_name=name),
        )

    @property
    def forms(self) -> list[Form]:
        """Return all the Form elements..

        Returns:
            list[Form]: A list of Form elements.
        """
        return cast(list[Form], self._filtered_elements("descendant::form:form"))


class Form(FormMixin, OfficeTargetFrameMixin):
    """Specification a user-interface form and defines the contents and
    properties of the form, "form:form".

    The "form:form" element is usable within the following elements:
    "form:form" and "office:forms".

    (The main objective of the current minimal implementation is to parse
    the existing form in a document.)
    """

    _tag = "form:form"
    _properties = (
        PropDef("name", "form:name"),
        PropDef("command", "form:command"),
        PropDef("datasource", "form:datasource"),
        PropDef("apply_filter", "form:apply-filter"),
        PropDef("command_type", "form:command-type"),
        PropDef("control_implementation", "form:control-implementation"),
    )

    def __init__(
        self,
        name: str | None = "Form",
        command: str | None = None,
        datasource: str | None = None,
        apply_filter: bool | None = None,
        command_type: str | None = None,
        control_implementation: str | None = None,
        target_frame: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a Form, "form:form".

        The "form:form" element is usable within the following elements:
        "form:form" and "office:forms".

        Args:
            name: The name of the form.
            command: The command for the form.
            datasource: The data source for the form.
            apply_filter: If True, a filter is applied.
            command_type: The type of the command.
            control_implementation: The control implementation.
            target_frame: The target frame for the form.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name is not None:
                self.name = name
            if command is not None:
                self.command = command
            if datasource is not None:
                self.datasource = datasource
            if apply_filter is not None:
                self.apply_filter = apply_filter
            if command_type is not None:
                self.command_type = command_type
            if control_implementation is not None:
                self.control_implementation = control_implementation
            if target_frame is not None:
                self.target_frame = target_frame

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name}>"


Form._define_attribut_property()

register_element_class(Form)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""Utility function to check for iterability.

This module provides a helper function to determine if an object is iterable,
with special handling for string and byte types.
"""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any


def isiterable(instance: Any) -> bool:
    """Checks if an object is iterable, excluding strings and bytes.

    This function considers strings and bytes objects as non-iterable, which is
    often the desired behavior when handling collections of items that might
    also include simple text.

    Args:
        instance: The object to check.

    Returns:
        bool: True if the object is iterable (and not a string/bytes),
            False otherwise.
    """
    if isinstance(instance, (str, bytes)):
        return False
    return isinstance(instance, Iterable)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.document import Document
from odfdo.paragraph import Span


@pytest.fixture
def body(samples) -> Iterable[Element]:
    document = Document(samples("span_style.odt"))
    yield document.body


def test_create_span():
    span = Span("my text", style="my_style")
    expected = '<text:span text:style-name="my_style">my text</text:span>'
    assert span.serialize() == expected


def test_create_span_naive_spaces():
    span = Span("my text   ", style="my_style")
    expected = (
        '<text:span text:style-name="my_style">my text<text:s text:c="3"/></text:span>'
    )
    assert span.serialize() == expected


def test_create_span_naive_spaces_no_format():
    span = Span("my text   ", style="my_style", formatted=False)
    expected = '<text:span text:style-name="my_style">my text </text:span>'
    assert span.serialize() == expected


def test_get_span_list(body):
    result = body.get_spans()
    assert len(result) == 2
    element = result[0]
    expected = '<text:span text:style-name="T1">moustache</text:span>'
    assert element.serialize() == expected


def test_get_span_list_property(body):
    result = body.spans
    assert len(result) == 2
    element = result[0]
    expected = '<text:span text:style-name="T1">moustache</text:span>'
    assert element.serialize() == expected


def test_get_span_list_style(body):
    result = body.get_spans(style="T2")
    assert len(result) == 1
    element = result[0]
    expected = '<text:span text:style-name="T2">rouge</text:span>'
    assert element.serialize() == expected


def test_get_span(body):
    span = body.get_span(position=1)
    expected = '<text:span text:style-name="T2">rouge</text:span>'
    assert span.serialize() == expected


def test_get_span_str(body):
    span = body.get_span(position=1)
    assert str(span) == "rouge"


def test_insert_span(body):
    span = Span("my text", "my_style")
    paragraph = body.get_paragraph(position=0)
    paragraph.append(span)
    read_span = paragraph.get_span(position=-1)
    assert read_span.serialize() == span.serialize()


def test_insert_span_line_break(body):
    span = Span("my text\nmultiline", "my_style")
    paragraph = body.get_paragraph(position=0)
    paragraph.append(span)
    read_span = paragraph.get_span(position=-1)
    assert read_span.serialize() == span.serialize()
    assert "text:line-break" in read_span.serialize()
    assert read_span.inner_text == "my text\nmultiline"


def test_insert_span_line_break_no_format(body):
    span = Span("my text\nmultiline", "my_style", formatted=False)
    paragraph = body.get_paragraph(position=0)
    paragraph.append(span)
    read_span = paragraph.get_span(position=-1)
    assert read_span.serialize() == span.serialize()
    assert "text:line-break" not in read_span.serialize()
    assert read_span.inner_text == "my text multiline"



#!/usr/bin/env python
"""Minimal example of copy of a style from another document.

Document.get_style() main parameters:
family        : The family of the style, text styles apply on individual
                characters.
display_name  : The name of the style as we see it in a desktop
                application. Styles have an internal name
                (“Yellow_20_Highlight” in this example) but here we use
                the display_name instead.
"""

import os
from pathlib import Path

from odfdo import Document

_DOC_SEQUENCE = 310
DATA = Path(__file__).parent / "data"
SOURCE = "lpod_styles.odt"


def generate_document() -> Document:
    """Return a document with a style read from another document."""
    document = Document("text")
    body = document.body
    body.clear()

    styled_source = Document(DATA / SOURCE)
    highlight_style = styled_source.get_style(
        family="text", display_name="Yellow Highlight"
    )

    document.insert_style(highlight_style, automatic=True)

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    style = document.get_style(family="text", display_name="Yellow Highlight")
    assert style.display_name == "Yellow Highlight"


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""Representation of the "styles.xml" part."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union, cast

from .element import Element
from .style_base import StyleBase
from .style_containers import OfficeAutomaticStyles, OfficeMasterStyles
from .utils import is_RFC3066
from .xmlpart import XmlPart

if TYPE_CHECKING:
    from .image import DrawFillImage, DrawMarker
    from .master_page import StyleMasterPage


CONTEXT_MAPPING = {
    "paragraph": ("//office:styles", "//office:automatic-styles"),
    "text": ("//office:styles", "//office:automatic-styles"),
    "graphic": ("//office:styles", "//office:automatic-styles"),
    "page-layout": ("//office:automatic-styles",),
    "master-page": ("//office:master-styles",),
    "font-face": ("//office:font-face-decls",),
    "outline": ("//office:styles", "//office:automatic-styles"),
    "date": ("//office:styles", "//office:automatic-styles"),
    "list": ("//office:styles", "//office:automatic-styles"),
    "presentation": ("//office:styles", "//office:automatic-styles"),
    "drawing-page": ("//office:automatic-styles",),
    "presentation-page-layout": ("//office:styles",),
    "marker": ("//office:styles",),
    "fill-image": ("//office:styles",),
    # FIXME Do they?
    "table": ("//office:styles", "//office:automatic-styles"),
    "table-cell": ("//office:styles", "//office:automatic-styles"),
    "table-row": ("//office:styles", "//office:automatic-styles"),
    "table-column": ("//office:styles", "//office:automatic-styles"),
    # FIXME: to test:
    "section": ("//office:styles", "//office:automatic-styles"),
    "chart": ("//office:styles", "//office:automatic-styles"),
}


class Styles(XmlPart):
    """Representation of the "styles.xml" part of an ODF document.

    This class provides an interface to the styles defined in the "styles.xml"
    part of an ODF document, which includes automatic styles, master styles,
    and default styles.
    """

    def _get_style_contexts(
        self, family: str, automatic: bool = False
    ) -> list[Element]:
        """Get the XML contexts where styles are stored.

        Args:
            family: The style family to search for.
            automatic: Whether to only search in automatic styles.

        Returns:
            list[Element]: A list of XML elements that are contexts for styles.
        """
        if automatic:
            elems = [self.get_element("//office:automatic-styles")]
        elif family:
            queries = CONTEXT_MAPPING.get(family) or (
                "//office:styles",
                "//office:automatic-styles",
            )
            elems = [self.get_element(query) for query in queries]
        else:
            # All possibilities
            elems = [
                self.get_element("//office:automatic-styles"),
                self.get_element("//office:styles"),
                self.get_element("//office:master-styles"),
                self.get_element("//office:font-face-decls"),
            ]
        return [e for e in elems if isinstance(e, Element)]

    def get_styles(
        self, family: str = "", automatic: bool = False
    ) -> list[StyleBase | DrawFillImage | DrawMarker]:
        """Return the list of styles in the Styles part.

        Optionally, the results can be limited to a specific family and/or
        to automatic styles.

        Args:
            family: The style family to filter by (e.g.,
                'paragraph', 'text').
            automatic: If True, only automatic styles are
                returned. Defaults to False.

        Returns:
            list[StyleBase|DrawFillImage|DrawMarker]: A list of style-like
            instances matching the criteria..
        """
        result = []
        for context in self._get_style_contexts(family, automatic=automatic):
            if context is None:  # pragma: nocover
                continue
            # print('-ctx----', automatic)
            # print(context.tag)
            # print(context.__class__)
            # print(context.serialize())
            result.extend(context.get_styles(family=family))
        return result

    @property
    def default_styles(self) -> list[StyleBase]:
        """Return the list of default styles ("style:default-style").

        Returns:
            list[StyleBase]: A list of default Style elements.
        """
        return cast(list[StyleBase], self.get_elements("//style:default-style"))

    def set_default_styles_language_country(self, value: str) -> None:
        """Set the default language and country in styles.

        Args:
            value: The language/country code in RFC3066 format (e.g., "en-US").

        Raises:
            TypeError: If the language code format is invalid.
        """
        language = str(value)
        if not is_RFC3066(language):
            msg = 'Language must be "xx" lang or "xx-YY" lang-COUNTRY code (RFC3066)'
            raise TypeError(msg)
        lc = language.split("-")
        if len(lc) == 2:
            lang, country = lc
        else:
            lang = lc[0]
            country = ""
        styles = [
            s for s in self.default_styles if s.family in {"graphic", "paragraph"}
        ]
        for style in styles:
            props = style.get_properties(area="text") or {}
            props["language"] = lang
            props["country"] = country
            style.set_properties(props, area="text")

    @property
    def default_language(self) -> str:
        """Get or set the default language from styles, in "en-US" format."""
        styles = [s for s in self.default_styles if s.family == "paragraph"]
        if not styles:
            return ""
        style = styles[0]
        props = style.get_properties(area="text") or {}
        lang = str(props.get("fo:language", ""))
        country = str(props.get("fo:country", ""))
        if lang and country:
            return f"{lang}-{country}"
        return lang or country

    @default_language.setter
    def default_language(self, value: str) -> None:
        return self.set_default_styles_language_country(value)

    def get_style(
        self,
        family: str,
        name_or_element: str | StyleBase | None = None,
        display_name: str | None = None,
    ) -> StyleBase | DrawFillImage | DrawMarker | None:
        """Return the style uniquely identified by its family and name.

        If the `name_or_element` argument is already a Style object, it will be
        returned. If `name_or_element` is None, the default style for the given
        family is fetched. If the name provided is a display name, use the
        `display_name` argument instead.

        Args:
            family: The style family (e.g., 'paragraph', 'text', 'graphic').
            name_or_element: The internal name of the
                style or a Style object itself.
            display_name: The display name of the style as seen in an
                office application.

        Returns:
            StyleBase | DrawFillImage | DrawMarker | None: A style-like
                instance, or None if not found.
        """
        for context in self._get_style_contexts(family):
            if context is None:
                continue  # pragma: nocover
            style = context.get_style(
                family,
                name_or_element=name_or_element,
                display_name=display_name,
            )
            if style is not None:
                return style
        return None

    @property
    def office_master_styles(self) -> OfficeMasterStyles | None:
        """Get or set the "office:master-styles" element.

        Returns:
            OfficeMasterStyles | None: The "office:master-styles" element, or None if not found.
        """
        return cast(
            Union[None, OfficeMasterStyles], self.get_element("//office:master-styles")
        )

    @office_master_styles.setter
    def office_master_styles(self, office_master_styles: OfficeMasterStyles) -> None:
        """Set the "office:master-styles" element.

        Args:
            office_master_styles: The "office:master-styles" element to set.
        """
        current = self.office_master_styles
        if isinstance(current, OfficeMasterStyles):
            current.delete()
        self.root.append(office_master_styles)

    @property
    def master_pages(self) -> list[StyleMasterPage]:
        """Get the list of master pages.

        Returns:
            list[StyleMasterPage]: A list of StyleMasterPage elements.
        """
        master_styles = self.office_master_styles
        if master_styles is None:
            return []
        return [
            e  # type: ignore[misc]
            for e in master_styles.children
            if isinstance(e, StyleBase) and e.tag == "style:master-page"
        ]

    def get_master_page(self, position: int = 0) -> StyleMasterPage | None:
        """Get a master page by its position.

        Args:
            position: The position (index) of the master page. Defaults to 0.

        Returns:
            StyleMasterPage | None: The StyleMasterPage element at the given position,
            or None if not found.
        """
        results = self.master_pages
        try:
            return results[position]
        except IndexError:
            return None

    @property
    def office_automatic_styles(self) -> OfficeAutomaticStyles | None:
        """Get or set the "office:automatic-styles" element.

        Returns:
            OfficeAutomaticStyles | None: The "office:automatic-styles" element,
            or None if not found.
        """
        return cast(
            Union[None, OfficeAutomaticStyles],
            self.get_element("//office:automatic-styles"),
        )

    @office_automatic_styles.setter
    def office_automatic_styles(
        self, office_automatic_styles: OfficeAutomaticStyles
    ) -> None:
        """Set the "office:automatic-styles" element.

        Args:
            office_automatic_styles: The "office:automatic-styles"
                element to set.
        """
        current = self.office_automatic_styles
        if isinstance(current, OfficeAutomaticStyles):
            current.delete()
        self.root.append(office_automatic_styles)



#!/usr/bin/env python
# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Script to check docstrings using griffe.

DocstringSectionParameters	"parameters"	Function parameters
DocstringSectionReturns	"returns"	Return values
DocstringSectionRaises	"raises"	Exceptions raised
DocstringSectionYields	"yields"	Generator yields
DocstringSectionReceives	"receives"	Generator receives
DocstringSectionExamples	"examples"	Usage examples
DocstringSectionAttributes	"attributes"	Class/module attributes
DocstringSectionFunctions	"functions"	Documented functions
DocstringSectionClasses	"classes"	Documented classes
DocstringSectionModules	"modules"	Documented modules
DocstringSectionTypeParameters	"typeParameters"	Type parameters
DocstringSectionOtherParameters	"otherParameters"	Keyword args
DocstringSectionWarns	"warns"	Warnings
DocstringSectionDeprecated	"deprecated"	Deprecation notice
DocstringSectionText	"text"	General text
DocstringSectionAdmonition	"admonition"	Admonitions

Available section kind values (all lowercase strings):
"text"
"parameters"
"other_parameters"
"returns"
"yields"
"receives"
"raises"
"warns"
"examples"
"attributes"
"functions"
"classes"
"modules"
"deprecated"
"admonition"
"typeParameters" (camelCase)

"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import griffe

MIN_DOC_LEN = 10
SPACER = 50
PACKAGE = "odfdo"
issues: list[str] = []


def add_issue(kind: str, obj: object, msg: str) -> None:
    issues.append(f"{kind} {obj.path:{SPACER}} - {msg}")


def print_issues() -> None:
    for issue in issues:
        print(issue)


def test_has_doc(fail: Callable, obj: Any) -> bool:
    if not obj.docstring:
        fail("missing docstring")
        return False
    return True


def get_sections(fail: Callable, obj: Any) -> Any:
    try:
        sections = obj.docstring.parse("auto")
    except Exception as e:
        fail(f"failed to parse docstring: {e}")
        return None
    return sections


def check_module(module: griffe.Module) -> None:
    def fail(msg: str) -> None:
        add_issue("O", module, msg)

    if not test_has_doc(fail, module):
        return

    get_sections(fail, module)
    # if not sections:
    #     return


def parse_modules(package: object) -> None:
    nb = len(package.modules)
    print(f"Parsing {nb} modules")
    for module in package.modules.values():
        check_module(module)


def check_class(klass: griffe.Class) -> None:
    def fail(msg: str) -> None:
        add_issue("C", klass, msg)

    if not test_has_doc(fail, klass):
        return

    sections = get_sections(fail, klass)
    if not sections:
        return

    for section in sections:
        if section.kind.value == "text" and len(section.value) < MIN_DOC_LEN:
            fail(f"description too short (<{MIN_DOC_LEN} char)")

    for member in klass.members.values():
        if member.is_function:  # Check if it's a method
            check_method(member)

    # class_obj.members	Dictionary of all class members (name → object)
    # member.is_function	Boolean check if member is a function/method
    # member.is_method	Alternative check for bound methods
    # member.docstring	The docstring object (if any)
    # member.docstring.value	Raw docstring text
    #
    # Instance methods (exclude static/class methods)
    # for name, member in class_obj.members.items():
    #     if member.is_function and not member.is_static and not member.is_classmethod:
    #         print(f"Instance method: {name}")

    # # Static methods only
    # for name, member in class_obj.members.items():
    #     if member.is_static:
    #         print(f"Static method: {name}")

    # # Class methods only
    # for name, member in class_obj.members.items():
    #     if member.is_classmethod:
    #         print(f"Class method: {name}")


def parse_classes(package: object) -> None:
    nb = len(package.classes)
    print(f"Parsing {nb} classes")
    for klass in package.classes.values():
        check_class(klass)


def check_method(meth: griffe.Method) -> None:
    def fail(msg: str) -> None:
        add_issue("E", meth, msg)

    if meth.name.startswith("_") and meth.name != "__init__":
        # do not test "_" methods
        return

    if not test_has_doc(fail, meth):
        return

    sections = get_sections(fail, meth)
    if not sections:
        return

    documented_params = set()
    has_returns = False
    has_yield = False

    for section in sections:
        if section.kind.value == "parameters":
            for param in section.value:
                name = param.name
                while name.startswith("*"):
                    name = name[1:]
                documented_params.add(name)
        elif section.kind.value == "returns":
            has_returns = True
        elif section.kind.value == "yields":
            has_yield = True

    actual_params: set = {
        p.name for p in meth.parameters if p.name not in ("self", "cls")
    }
    undocumented: set = actual_params - documented_params
    not_existing: set = documented_params - actual_params

    undocumented -= {"kwargs", "args"}

    if undocumented:
        fail(f"undocumented parameters: {', '.join(undocumented)}")
    if not_existing:
        fail(f"documented parameters not found: {', '.join(not_existing)}")
    if meth.returns and not has_returns and meth.returns != "None" and not has_yield:
        fail(f"has return type ({meth.returns}) but no return documentation")


def check_function(func: griffe.Function) -> None:
    def fail(msg: str) -> None:
        add_issue("F", func, msg)

    if not test_has_doc(fail, func):
        return

    sections = get_sections(fail, func)
    if not sections:
        return

    documented_params = set()
    has_returns = False

    for section in sections:
        if section.kind.value == "parameters":
            for param in section.value:
                documented_params.add(param.name)
        elif section.kind.value == "returns":
            has_returns = True

    actual_params = {p.name for p in func.parameters if p.name not in ("self", "cls")}
    undocumented = actual_params - documented_params
    not_existing = documented_params - actual_params

    if undocumented:
        fail(f"undocumented parameters: {', '.join(undocumented)}")
    if not_existing:
        fail(f"documented parameters not found: {', '.join(not_existing)}")
    if func.returns and not has_returns:
        fail("has return type but no return documentation")


def parse_functions(package: object) -> None:
    nb = len(package.functions)
    print(f"Parsing {nb} functions")
    for func in package.functions.values():
        check_function(func)


def main() -> None:
    package = griffe.load(PACKAGE)
    parse_modules(package)
    parse_classes(package)
    parse_functions(package)
    print_issues()


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest

from odfdo.scripts import diff
from odfdo.scripts.diff import main as main_script
from odfdo.scripts.diff import main_diff, parse_cli_args

SCRIPT = Path(diff.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_diff_no_param():
    params = []
    _out, err, exitcode = run_params(params)
    assert exitcode == 2
    assert "odfdo-diff: error" in err
    assert "usage" in err


# direct access to internal function


def test_diff_2_no_param_on_main_function(monkeypatch, capsys):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.type is SystemExit
        assert result.value.code >= 1
    captured = capsys.readouterr()

    assert "usage" in captured.err


def test_diff_2_no_param(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args([])
        assert result.type is SystemExit
        assert result.value.code >= 1
    captured = capsys.readouterr()

    assert "usage" in captured.err


def test_diff_2_no_file(capsys):
    params = parse_cli_args(["none_file1", "none_file2"])

    with pytest.raises(SystemExit) as result:
        main_diff(params)
        assert result.type is SystemExit
        assert result.value.code == 1
    captured = capsys.readouterr()

    assert "Show a diff" in captured.out


def test_diff_2_no_file_on_main_function(monkeypatch, capsys):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", ["odfdo-diff", "none_file1", "none_file2"])
        main_script()
        assert result.type is SystemExit
        assert result.value.code == 1
    captured = capsys.readouterr()

    assert "Show a diff" in captured.out


def test_diff_2_bad_format(capsys, samples):
    params = parse_cli_args(
        [f"{samples('test_diff1.odt')}", f"{samples('background.odp')}"]
    )

    with pytest.raises(SystemExit) as result:
        main_diff(params)
        assert result.type is SystemExit
        assert result.value.code == 1
    captured = capsys.readouterr()

    assert "Show a diff" in captured.out
    assert "Error: odfdo-diff requires input documents of type text" in captured.out


def test_diff_2_diff(capsys, samples):
    params = parse_cli_args(
        [f"{samples('test_diff1.odt')}", f"{samples('test_diff2.odt')}"]
    )

    main_diff(params)
    captured = capsys.readouterr()

    assert "test_diff1.odt" in captured.out
    assert "test_diff2.odt" in captured.out
    assert "\n-content A\n" in captured.out
    assert "\n+content B\n" in captured.out


def test_diff_2_ndiff(capsys, samples):
    params = parse_cli_args(
        ["-n", f"{samples('test_diff1.odt')}", f"{samples('test_diff2.odt')}"]
    )

    main_diff(params)
    captured = capsys.readouterr()

    assert "test_diff1.odt" not in captured.out
    assert "\n- content A\n" in captured.out
    assert "\n+ content B\n" in captured.out


def test_diff_2_diff_same(capsys, samples):
    params = parse_cli_args(
        [f"{samples('test_diff1.odt')}", f"{samples('test_diff1.odt')}"]
    )

    main_diff(params)
    captured = capsys.readouterr()

    assert not captured.out.strip()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

from odfdo.table import Cell
from odfdo.user_field import UserFieldDecl
from odfdo.variable import VarSet

ZOE = "你好 Zoé"


def test_get_with_cell():
    cell = Cell(42)
    assert cell.get_value() == 42


def test_get_with_variable():
    variable_set = VarSet(ZOE, 42)
    assert variable_set.get_value() == 42


def test_get_with_user_field():
    user_field_decl = UserFieldDecl(ZOE, 42)
    assert user_field_decl.get_value() == 42


def test_with_cell():
    cell = Cell(42)
    cell.set_value(ZOE)
    expected = (
        '<table:table-cell office:value-type="string" '
        'calcext:value-type="string" '
        f'office:string-value="{ZOE}">'
        "<text:p>"
        f"{ZOE}"
        "</text:p>"
        "</table:table-cell>"
    )
    assert cell.serialize() == expected


def test_with_variable():
    variable_set = VarSet(ZOE, 42)
    variable_set.set_value(ZOE)
    expected = (
        '<text:variable-set office:value-type="string" '
        'calcext:value-type="string" '
        f'office:string-value="{ZOE}" text:name="{ZOE}" '
        'text:display="none">'
        f"{ZOE}"
        "</text:variable-set>"
    )
    assert variable_set.serialize() == expected


def test_with_user_field():
    user_field_decl = UserFieldDecl(ZOE, 42)
    user_field_decl.set_value(ZOE)
    expected = (
        '<text:user-field-decl office:value-type="string" '
        'calcext:value-type="string" '
        'office:string-value="%s" text:name="%s"/>'
    ) % ((ZOE,) * 2)
    assert user_field_decl.serialize() == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormButton


@pytest.fixture
def form_button() -> Iterable[FormButton]:
    yield FormButton(
        name="some name",
        title="some title",
        label="some label",
        button_type="push",
        default_button=True,
        repeat=True,
        focus_on_click=True,
        image_align="start",
        image_position="top",
        image_data="data",
        toggle=True,
        value="special",
        target_frame="some target",
        href="some url",
        control_implementation="some implem",
        disabled=False,
        printable=True,
        tab_index=4,
        tab_stop=False,
        xml_id="control1",
    )


def test_form_button_class():
    form = FormButton()
    assert isinstance(form, FormButton)


def test_form_button_minimal_tag():
    form = Element.from_tag("<form:button/>")
    assert isinstance(form, FormButton)


def test_form_button_serialize():
    form = FormButton()
    assert form.serialize() == "<form:button/>"


def test_form_button_repr():
    form = FormButton()
    assert repr(form) == "<FormButton name=None xml_id=None>"


def test_form_button_name(form_button):
    assert form_button.name == "some name"


def test_form_button_title(form_button):
    assert form_button.title == "some title"


def test_form_button_label(form_button):
    assert form_button.label == "some label"


def test_form_button_button_type_1(form_button):
    assert form_button.button_type == "push"


def test_form_button_button_type_2(form_button):
    with pytest.raises(ValueError):
        form_button.button_type = "bad"


def test_form_button_button_type_3(form_button):
    form_button.button_type = "reset"
    assert form_button.button_type == "reset"


def test_form_button_default_button(form_button):
    assert form_button.default_button is True


def test_form_button_repeat(form_button):
    assert form_button.repeat is True


def test_form_button_delay_for_repeat_1(form_button):
    assert form_button.delay_for_repeat == "PT0.050S"


def test_form_button_delay_for_repeat_2(form_button):
    form_button.delay_for_repeat = "PT50S"
    assert form_button.delay_for_repeat == "PT50S"


def test_form_button_delay_for_repeat_3():
    form = FormButton(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        delay_for_repeat="PT50S",
    )
    assert form.delay_for_repeat == "PT50S"


def test_form_button_focus_on_click(form_button):
    assert form_button.focus_on_click is True


def test_form_button_image_align_1(form_button):
    assert form_button.image_align == "start"


def test_form_button_image_align_2(form_button):
    with pytest.raises(ValueError):
        form_button.image_align = "bad"


def test_form_button_image_align_3(form_button):
    form_button.image_align = "end"
    assert form_button.image_align == "end"


def test_form_button_image_position_1(form_button):
    assert form_button.image_position == "top"


def test_form_button_image_position_2(form_button):
    with pytest.raises(ValueError):
        form_button.image_position = "bad"


def test_form_button_image_position_3(form_button):
    form_button.image_position = "bottom"
    assert form_button.image_position == "bottom"


def test_form_button_image_data(form_button):
    assert form_button.image_data == "data"


def test_form_button_toggle(form_button):
    assert form_button.toggle is True


def test_form_button_value(form_button):
    assert form_button.value == "special"


def test_form_button_target_frame(form_button):
    assert form_button.target_frame == "some target"


def test_form_button_href(form_button):
    assert form_button.href == "some url"


def test_form_button_control_implementation(form_button):
    assert form_button.control_implementation == "some implem"


def test_form_button_disabled(form_button):
    assert form_button.disabled is False


def test_form_button_printable(form_button):
    assert form_button.printable is True


def test_form_button_tab_index(form_button):
    assert form_button.tab_index == 4


def test_form_button_tab_stop(form_button):
    assert form_button.tab_stop is False


def test_form_button_xml_id(form_button):
    assert form_button.xml_id == "control1"


def test_form_button_form_id(form_button):
    assert form_button.form_id == "control1"


def test_form_button_xforms_submission_1(form_button):
    assert form_button.xforms_submission is None


def test_form_button_xforms_submission_2():
    form = FormButton(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        xforms_submission="something",
    )
    assert form.xforms_submission == "something"


def test_form_button_form_id_2():
    form = FormButton(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"
